import polars as pl

__version__: str

def get_crs_from_code(srid: int) -> str | None: ...
def get_crs_authority(definition: str) -> tuple[str, str] | None: ...
def to_python_dict(series: pl.Series) -> pl.Series: ...
